package main

import (
	"context"
	"crypto/tls"
	"fmt"
	"io"
	"log"
	"net"
	"os"
	"strings"
	"sync"

	"golang.org/x/crypto/acme/autocert"
)

func usage() {
	fmt.Println(`
Usage:
  achmed [OPTIONS]...

Options:
  --bind [HOST]:PORT   Change bind address (default: ":443")
                        Specify multiple times to add additional bind addresses.
  --cache-dir PATH     Directory for certificate cache (default: .)
                        The directory will be created if it doesn't exist.
  DNSNAME=UPSTREAM     Listen on DNSNAME, forward connections to tcp:UPSTREAM
  --help               Display this message
`)
	os.Exit(0)
}

type AchmedServer struct {
	pairs                  map[string]string
	bindAddrs              []string
	anyCustomBindAddresses bool
	cacheDir               string
	errInvalidHost         error
}

func NewAchmedServer(args []string) *AchmedServer {
	this := &AchmedServer{
		pairs:                  make(map[string]string),
		bindAddrs:              []string{":443"},
		anyCustomBindAddresses: false,
		cacheDir:               ".",
		errInvalidHost:         fmt.Errorf("Invalid host"),
	}

	for i := 0; i < len(args); i += 1 {
		arg := args[i]
		if arg == "--help" {
			usage()

		} else if arg == "--bind" && i+1 < len(args) {
			i += 1
			if !this.anyCustomBindAddresses {
				this.bindAddrs = []string{}
			}
			this.bindAddrs = append(this.bindAddrs, args[i])

		} else if arg == "--cache-dir" && i+1 < len(args) {
			i += 1
			this.cacheDir = args[i]

		} else if parts := strings.Split(arg, "="); len(parts) == 2 {
			this.pairs[strings.ToLower(parts[0])] = parts[1]

		} else {
			fmt.Fprintf(os.Stderr, "Unknown argument '%s' (try --help)\n", arg)
			os.Exit(1)
		}
	}

	return this
}

func (this *AchmedServer) HostPolicy(ctx context.Context, domain string) error {
	if _, ok := this.pairs[domain]; !ok {
		return this.errInvalidHost
	}

	return nil
}

func (this *AchmedServer) Start() {

	amgr := autocert.Manager{
		Prompt:     autocert.AcceptTOS,
		Cache:      autocert.DirCache(this.cacheDir),
		HostPolicy: this.HostPolicy,
	}

	tlsConfig := tls.Config{
		GetCertificate:           amgr.GetCertificate,
		PreferServerCipherSuites: true,
	}

	// Listen on all addresses
	for _, bindAddr := range this.bindAddrs {
		go this.Listen(bindAddr, &tlsConfig)
	}

	// Block forever
	select {}
}

func (this *AchmedServer) Listen(bindAddr string, tlsConfig *tls.Config) {
	log.Printf("Listening on %s...", bindAddr)

	l, err := tls.Listen("tcp", bindAddr, tlsConfig)
	if err != nil {
		log.Fatalf("Listen('%s'): %s", bindAddr, err.Error())
	}

	defer l.Close()

	for {
		conn, err := l.Accept()
		if err != nil {
			log.Printf("Accept('%s'): %s", bindAddr, err.Error())
			continue
		}

		go this.HandleConnection(conn)
	}
}

func (this *AchmedServer) HandleConnection(netConn net.Conn) {
	defer netConn.Close()

	tlsConn, ok := netConn.(*tls.Conn)
	if !ok {
		log.Printf("%s: not a TLS connection", netConn.RemoteAddr().String())
		return
	}

	sni := tlsConn.ConnectionState().ServerName
	upstreamAddress, ok := this.pairs[strings.ToLower(sni)]
	if !ok {
		log.Printf("%s: no hostname match", netConn.RemoteAddr().String())
		return
	}

	// Try and make upstream connection
	upstreamConn, err := net.Dial("tcp", upstreamAddress)
	if err != nil {
		log.Printf("%s: couldn't connect to upstream '%s': %s", netConn.RemoteAddr().String(), upstreamAddress, err.Error())
		return
	}

	defer upstreamConn.Close()

	// Proxy data until both read/write channels close
	wg := sync.WaitGroup{}
	wg.Add(2)

	go func() {
		defer wg.Done()
		io.Copy(tlsConn, upstreamConn)
	}()
	go func() {
		defer wg.Done()
		io.Copy(upstreamConn, tlsConn)
	}()

	wg.Wait()
}

func main() {
	as := NewAchmedServer(os.Args[1:])
	as.Start()
}
